@@ -1,30 +1,33 @@
 {-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE FlexibleContexts #-}
 -- TODO: suggest the convenience functions be put into Hint proper?
 module Mueval.Interpreter where
 
-import Control.Monad (forM_,guard,mplus,unless,when)
-import Control.Monad.Trans (MonadIO)
-import Control.Monad.Writer (Any(..),runWriterT,tell)
-import Data.Char (isDigit)
-import Data.List (stripPrefix)
-import System.Directory (copyFile, makeRelativeToCurrentDirectory, removeFile, setCurrentDirectory)
-import System.Exit (exitFailure)
-import System.FilePath.Posix (takeFileName)
 import qualified Control.Exception.Extensible as E (evaluate,catch,SomeException(..))
+import           Control.Monad (forM_,guard,mplus,unless,when)
+import           Control.Monad.Trans (MonadIO)
+import           Control.Monad.Writer (Any(..),runWriterT,tell)
+import           Data.Char (isDigit)
 
-import qualified System.IO.UTF8 as UTF (putStrLn)
+import           System.Directory
 
-import Language.Haskell.Interpreter (eval, set, reset, setImportsQ, loadModules, liftIO,
+import           System.Exit (exitFailure)
+import           System.FilePath.Posix (takeBaseName)
+import           System.IO (openTempFile)
+
+import           Data.List
+
+import           Language.Haskell.Interpreter (eval, set, reset, setImportsQ, loadModules, liftIO,
                                      installedModulesInScope, languageExtensions,
                                      typeOf, setTopLevelModules, runInterpreter, glasgowExtensions,
                                      OptionVal(..), Interpreter,
                                      InterpreterError(..),GhcError(..),
                                      Extension(UnknownExtension))
-import Language.Haskell.Interpreter.Unsafe (unsafeSetGhcOption)
+import           Language.Haskell.Interpreter.Unsafe (unsafeSetGhcOption)
 
-import Mueval.ArgsParse (Options(..))
+import           Mueval.ArgsParse (Options(..))
 import qualified Mueval.Resources as MR (limitResources)
-import qualified Mueval.Context  as MC (qualifiedModules)
+import qualified Mueval.Context as MC (qualifiedModules)
 
 readExt :: String -> Extension
 readExt s = case reads s of
@@ -39,6 +42,7 @@ readExt s = case reads s of
 interpreter :: Options -> Interpreter (String,String,String)
 interpreter Options { extensions = exts, namedExtensions = nexts,
                       rLimits = rlimits,
+                      typeOnly = noEval,
                       loadFile = load, expression = expr,
                       packageTrust = trust,
                       trustedPackages = trustPkgs,
@@ -54,12 +58,17 @@ interpreter Options { extensions = exts, namedExtensions = nexts,
 
                                   reset -- Make sure nothing is available
                                   set [installedModulesInScope := False]
-                                  let lfl' = takeFileName load
-                                  when (load /= "") $ do liftIO (mvload load)
-                                                         loadModules [lfl']
-                                                         -- We need to mangle the String to
-                                                         -- turn a filename into a module.
-                                                         setTopLevelModules [takeWhile (/='.') lfl']
+
+                                  -- if we're given a file of definitions, we need to first copy it to a temporary file in /tmp (cpload),
+                                  -- then tell Hint to parse/read it, then extract the 'module name' of the file,
+                                  -- and tell Hint to expose the module into memory; then we need to store the temporary file's filepath
+                                  -- so we can try to clean up after ourselves later.
+                                  lfl' <- if (load /= "") then (do { lfl <- liftIO (cpload load);
+                                                                     loadModules [lfl];
+                                                                     -- We need to mangle the String to
+                                                                     -- turn a filename into a module.
+                                                                     setTopLevelModules [takeBaseName load];
+                                                                     return lfl }) else (return "")
 
                                   liftIO $ MR.limitResources rlimits
 
@@ -68,13 +77,15 @@ interpreter Options { extensions = exts, namedExtensions = nexts,
                                     Just ms -> do let unqualModules =  zip ms (repeat Nothing)
                                                   setImportsQ (unqualModules ++ MC.qualifiedModules)
 
-                                  -- clean up our tmp file here; must be after setImportsQ
+                                  -- clean up our tmp file here; must be *after* setImportsQ
                                   when (load /= "") $ liftIO (removeFile lfl')
 
-                                  -- we don't check if the expression typechecks
+                                  -- we don't deliberately don't check if the expression typechecks
                                   -- this way we get an "InterpreterError" we can display
                                   etype <- typeOf expr
-                                  result <- eval expr
+                                  result <- if noEval
+                                               then return ""
+                                               else eval expr
 
                                   return (expr, etype, result)
 
@@ -84,12 +95,19 @@ interpreterSession :: Options -> IO ()
 interpreterSession opts = do r <- runInterpreter (interpreter opts)
                              case r of
                                  Left err -> printInterpreterError err
-                                 Right (e,et,val) -> when (printType opts) (sayIO e >> sayIO et) >> sayIO val
+                                 Right (e,et,val) -> do when (printType opts)
+                                                             (sayIO e >> sayIOOneLine et)
+                                                        sayIO val
+  where sayIOOneLine = sayIO . unwords . words
 
-mvload :: FilePath -> IO ()
-mvload lfl = do canonfile <- makeRelativeToCurrentDirectory lfl
-                liftIO $ copyFile canonfile $ "/tmp/" ++ takeFileName canonfile
-                setCurrentDirectory "/tmp" -- will at least mess up relative links
+-- | Given a filepath (containing function definitions), copy it to a temporary file and change directory to it, returning the new filepath.
+cpload :: FilePath -> IO FilePath
+cpload definitions = do
+                tmpdir <- getTemporaryDirectory
+                (tempfile,_) <- System.IO.openTempFile tmpdir "mueval.hs"
+                liftIO $ copyFile definitions tempfile
+                setCurrentDirectory tmpdir -- will at least mess up relative links
+                return tempfile
 
 ---------------------------------
 -- Handling and outputting results
@@ -100,7 +118,7 @@ mvload lfl = do canonfile <- makeRelativeToCurrentDirectory lfl
 -- flooding. Lambdabot has a similar limit.
 sayIO :: String -> IO ()
 sayIO str = do (out,b) <- render 1024 str
-               UTF.putStrLn out
+               putStrLn out
                when b exitFailure
 
 -- | Oh no, something has gone wrong. If it's a compilation error pretty print
